//////////////////////////////////////////////////////////////////////
// IOCntrl.cpp: implementation of the CIOCntrl class.
//
//////////////////////////////////////////////////////////////////////
// Copyright 2000.    Moe Wheatley AE4JY  <ae4jy@qsl.net>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
////////////////////////////////////////////////////////////////////////
//

#include <float.h>
#include <afxmt.h>
#include "stdafx.h"
#include "Errorcodes.h"
#include "IOCntrl.h"
#include "perform.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define BUF_SIZE 2048	//size of data chunks to process at a time from
						// the soundcard
#define MAXTIMER 10		//wait for 1 seconds for thread to start
#define SAMP_RATE 8000
#define RMSAVE 10

double DebugFP1=0.0;
double DebugFP2=0.0;
INT DebugINT=0x0;

double CoreData[512];

// Non Class routine for launching the worker thread

static UINT ThreadLauncher( CIOCntrl* pIO)
{
	pIO->ProcessLoop();
	return 1;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CIOCntrl::CIOCntrl()
{
	m_DLLRevision = 133;	//DLL revision number.(100 = 1.00)
	m_fProcThreadQuit = FALSE;
	m_DLLStatus = DLLSTAT_OFF;
	m_pProcThread = NULL;
	m_ErrorMessage = _T("");
	m_ThreadActive = FALSE;
	m_hWnd = NULL;
	m_UseInWave = FALSE;
	m_UseOutWave = FALSE;
	m_BufferRdy = FALSE;
	m_UseAFC = TRUE;
	m_pSoundDev = NULL;
	m_pBuffer = NULL;
	m_pCfft = NULL;
	m_pBpskdet = NULL;
	m_LastRcvState = DLLSTAT_OFF;
	m_sOutFilePath = "";
	m_OutFileLimit = 480000;
	m_WaveOutPaused = FALSE;
	m_AppendMode = FALSE;
	m_WaveInOffset = 0;
	InitializeCriticalSection(&m_CriticalSection);
}

CIOCntrl::~CIOCntrl()
{
	StopIO();
	DeleteResources();
	DeleteCriticalSection(&m_CriticalSection);
	if(m_pCfft)
	{
		delete m_pCfft;	
		m_pCfft = NULL;
	}
	if(m_pBpskdet)
	{
		delete m_pBpskdet;
		m_pBpskdet = NULL;
	}
}

//*************************************************************************
//**************  routines exposed to the outside  ************************
//*************************************************************************

///////////////////////////////////////////////////////////////////////////
//  Create all the objects and initialize them prior to the worker thread
//       starting.
///////////////////////////////////////////////////////////////////////////
BOOL CIOCntrl::InitSystem()
{
// Setup Soundcard values
InitPerformance();
	m_WaveFormatEx.wFormatTag = WAVE_FORMAT_PCM;
	m_WaveFormatEx.nChannels = 1;
	m_WaveFormatEx.wBitsPerSample = 16;
	m_WaveFormatEx.nSamplesPerSec = SAMP_RATE;
	m_WaveFormatEx.nBlockAlign = 
			m_WaveFormatEx.nChannels *(m_WaveFormatEx.wBitsPerSample/8);
	m_WaveFormatEx.nAvgBytesPerSec = 
			m_WaveFormatEx.nSamplesPerSec *
			m_WaveFormatEx.nBlockAlign;
	m_WaveFormatEx.cbSize = 0;

	m_pSoundDev = new CSound;	//create sound object
	ASSERT( m_pSoundDev);
	if(!m_pSoundDev)
		return FALSE;
	if( !m_pCfft )
		m_pCfft = new Cfft;			//create fft object
	ASSERT( m_pCfft);
	if(!m_pCfft)
		return FALSE;
	if(!m_pBpskdet)
		m_pBpskdet = new CBPSKDet;
	ASSERT( m_pBpskdet);
	if(!m_pBpskdet)
		return FALSE;
//allocate buffer for soundcard/wavefile samples
	if( (m_pBuffer = new double[sizeof(double)*BUF_SIZE]) == NULL )
		return FALSE;
	return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::StopIO()
{
	if( m_pProcThread != NULL )		//if thread is already running
	{
		m_ThreadActive = FALSE;
		m_fProcThreadQuit = TRUE;	// stop it
		::WaitForSingleObject(m_pProcThread->m_hThread,	1500);
		delete m_pProcThread;
		m_pProcThread = NULL;
		m_DLLStatus = DLLSTAT_OFF;
//ReadPerformance();
	}
}


///////////////////////////////////////////////////////////////////////////
//   Called to start the I/O thread running
///////////////////////////////////////////////////////////////////////////
INT CIOCntrl::StartIO(HWND h_Wnd, BOOL UseInWave, BOOL UseOutWave, BOOL UseAFC)
{
INT Timer;
	m_UseInWave = UseInWave;
	m_UseOutWave = UseOutWave;
	m_UseAFC = UseAFC;
	m_hWnd = h_Wnd;
	if( m_pProcThread == NULL)
	{
		if(	!m_pSoundDev || !m_pCfft)	//need to re instantiate?
		{
			if( !InitSystem() )
				return DLLERR_MEMORY;
		}
		m_ThreadActive = FALSE;
		m_DLLStatus = DLLSTAT_OFF;
		m_fProcThreadQuit = FALSE;
		m_pProcThread = AfxBeginThread(
			(AFX_THREADPROC)ThreadLauncher,	//thread function
			(LPVOID)this,					// ptr to this class
			THREAD_PRIORITY_NORMAL, 		//give worker thread priority
			0,								// same stack size
			CREATE_SUSPENDED,				// don't let it start yet
			NULL );							// same security attributes
		if(!m_pProcThread)
			return DLLSTAT_OFF;
		m_pProcThread->m_bAutoDelete = FALSE;	// keep thread object around
		m_pProcThread->ResumeThread ();	//let er rip
		Timer = MAXTIMER;
		while( (m_DLLStatus == DLLSTAT_OFF) && !m_fProcThreadQuit && (Timer-- > 0) )
			::Sleep(100);
	}
	::Sleep(100);
	if(m_DLLStatus >= 0)
		return m_DLLStatus;
	else
	{
		if(m_DLLErrStatus > 0)
			return m_DLLErrStatus;
		else
			return DLLERR_NONE;
	}
}


///////////////////////////////////////////////////////////////////////////
// Initializes the FFT 
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::SetFFTParams(INT ave,double gain, INT type )
{
	if(m_pCfft)
		m_pCfft->SetFFTParams( ave, gain, type );
}

///////////////////////////////////////////////////////////////////////////
//  checks/Sets the wave file input path
///////////////////////////////////////////////////////////////////////////
INT CIOCntrl::SetInFilePath(CString sPath, long* FileSize, INT Offset)
{
INT ErrorCode;
	m_sInFilePath = sPath;
	if( Offset >= 0 )
		m_WaveInOffset = Offset;
	ErrorCode = WaveDev.InOpen( &m_sInFilePath, &m_WaveFormatEx,
						0, &m_FileSize, 0);
	if( (m_WaveFormatEx.nChannels != 1) ||
		( m_WaveFormatEx.nSamplesPerSec < (SAMP_RATE-150) )  ||
		( m_WaveFormatEx.nSamplesPerSec > (SAMP_RATE+150) )  ||
		( m_FileSize == 0) )
	{
		ErrorCode = WAVERR_BADFORMAT;
		ProcessError( ErrorCode );
		return ErrorCode;
	}
	*FileSize = m_FileSize;
	return ErrorCode;
}

///////////////////////////////////////////////////////////////////////////
//  checks/Sets the wave file output path
///////////////////////////////////////////////////////////////////////////
INT CIOCntrl::SetOutWavePath(CString sPath, BOOL AppendMode)
{
LONGLONG freespace;
LONGLONG tmp;
	m_sOutFilePath = sPath;
	m_AppendMode = AppendMode;
	freespace = WaveDev.GetDriveFreeSpace(&m_sOutFilePath);
	if(freespace !=0 )
	{
		tmp = ((9*freespace)/10);
		if(tmp > (8*60*60*16000) )	//clamp to 8 hours
			m_OutFileLimit = (8*60*60*16000);
		else
			m_OutFileLimit = (int)tmp;
	}
	else
		m_OutFileLimit = 0;
	return m_OutFileLimit;
}

///////////////////////////////////////////////////////////////////////////
// Gets latest raw data and put's it in users array
///////////////////////////////////////////////////////////////////////////
BOOL CIOCntrl::GetFFTData(LONG* pData, LONG start, LONG end)
{
BOOL Ovrld = FALSE;
	if( !m_fProcThreadQuit && (m_pCfft != NULL) && m_BufferRdy)
	{
		EnterCriticalSection(&m_CriticalSection);
		Ovrld = m_pCfft->GetFFTData( start, end, pData);
		LeaveCriticalSection(&m_CriticalSection);
	}
	else
	{
		for(INT i=start; i<=end; i++)
			pData[i] = 0;
	}
	m_BufferRdy = FALSE;
	return Ovrld;
}

///////////////////////////////////////////////////////////////////////////
//  Resets the radio frequency control parameters if not decoding
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::ResetRigCtrl()
{
	if( m_pBpskdet->GetDecoderState() == DLLSTAT_RXCRC_NO_SIGNAL)
		m_RigCntrl.ResetCtrl(m_pBpskdet);
}

//*************************************************************************
//**********************  local routines  *********************************
//*************************************************************************

///////////////////////////////////////////////////////////////////////////
//  Main I/O thread processing loop
// Process for Receiving audio data from soundcard
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::ProcessLoop()
{
INT result;
INT ErrorCode;
	Sleep(50);
	m_BufferRdy = FALSE;
	m_RigCntrl.ResetCtrl(m_pBpskdet);
	if(m_UseInWave)
	{
		ErrorCode = WaveDev.InOpen( &m_sInFilePath, &m_WaveFormatEx,
							BUF_SIZE, &m_FileSize, m_WaveInOffset*SAMP_RATE);
		if( (m_WaveFormatEx.nChannels != 1) ||
			( m_WaveFormatEx.nSamplesPerSec < (SAMP_RATE-150) )  ||
			( m_WaveFormatEx.nSamplesPerSec > (SAMP_RATE+150) )  ||
			( m_FileSize == 0) )
		{
			ErrorCode = WAVERR_BADFORMAT;
			ProcessError( ErrorCode );
		}
		m_UseOutWave = FALSE;
	}
	else
	{
		m_WaveFormatEx.nSamplesPerSec = SAMP_RATE;
		m_WaveFormatEx.nAvgBytesPerSec = 
				m_WaveFormatEx.nSamplesPerSec *
				m_WaveFormatEx.nBlockAlign;
		ErrorCode = m_pSoundDev->InOpen( &m_WaveFormatEx, BUF_SIZE,
											0, m_CardNum);
		if(m_UseOutWave && (ErrorCode==NO_ERROR) )
		{
			ErrorCode = WaveDev.OutOpen( &m_sOutFilePath, &m_WaveFormatEx,
						BUF_SIZE, m_OutFileLimit/2, !m_AppendMode);
		}
	}
	if( ErrorCode != NO_ERROR )
	{
		ProcessError( ErrorCode );
		StopThread();
	}
	if(m_pCfft)
		m_pCfft->ResetFFT();
	if(m_pBpskdet)
	{
		if(m_UseInWave)
			m_pBpskdet->Init(m_WaveFormatEx.nSamplesPerSec, BUF_SIZE, m_hWnd);
		else
			m_pBpskdet->Init(SAMP_RATE, BUF_SIZE, m_hWnd);
	}
	m_ThreadActive = TRUE;
	StatusChange(DLLSTAT_RXCRC_NO_SIGNAL);

	Sleep(500);
	m_ProgressCount = 0;
	m_Progress = 0;
	while( !m_fProcThreadQuit )
	{
		if(m_UseInWave)
		{
			if( (ErrorCode=WaveDev.InRead( m_pBuffer, BUF_SIZE )) != BUF_SIZE)
			{
				m_BufferRdy = FALSE;
				if( ErrorCode < 0)
				{
					ErrorCode = WaveDev.GetError();
					ProcessError( ErrorCode );
				}
				else
				{
					ProcessError( DLLSTAT_WAVEFINISHED );
				}
				WaveDev.InClose();
				StopThread();
			}
			m_ProgressCount += BUF_SIZE;
			m_Progress = 1 + (100*m_ProgressCount)/m_FileSize;
			Sleep(25);
		}
		else
		{
			if( (result = m_pSoundDev->InRead( m_pBuffer, BUF_SIZE )) != BUF_SIZE)
			{
				ErrorCode = m_pSoundDev->GetError();
				if( (ErrorCode == SOUNDIN_ERR_OVERFLOW) ||
					(ErrorCode == SOUNDIN_ERR_TIMEOUT) )
				{								//cpu couldn't keep up
												// so try again
					m_pSoundDev->InClose();
					StatusChange(DLLSTAT_CPUSLOW);
					ErrorCode = m_pSoundDev->InOpen( &m_WaveFormatEx, BUF_SIZE, 0, m_CardNum);
				}
				if( ErrorCode != NO_ERRORS)	//must be some other soundcard error
				{
					m_pSoundDev->InClose();
					ProcessError( ErrorCode );
					StopThread();
				}
			}
			else
			{
				if( m_UseOutWave )
				{
					if(!m_WaveOutPaused)
					{
						ErrorCode = WaveDev.OutWrite( m_pBuffer, BUF_SIZE);
						if( ErrorCode != BUF_SIZE)
						{
							WaveDev.OutClose();
							if( ErrorCode != NO_ERRORS)
								ProcessError( ErrorCode );
							else
								ProcessError( DLLSTAT_WAVEFINISHED );
							StopThread();
						}
						m_ProgressCount += BUF_SIZE;
						m_Progress =  (100*m_ProgressCount)/(m_OutFileLimit/2);
					}
				}
				else
					m_Progress = 0;

			}
		}
		EnterCriticalSection(&m_CriticalSection);
//Calculate 1024 pt. FFT on both halves of the 2048 sample buffer
//StartPerformance();
		m_pCfft->CalcFFT( m_pBuffer );
		m_pCfft->CalcFFT( m_pBuffer+1024 );
//StopPerformance();
		m_BufferRdy = TRUE;
		LeaveCriticalSection(&m_CriticalSection);
		::PostMessage( m_hWnd, MSG_FFTDATARDY, m_Progress, m_pBpskdet->GetSNR() );

// Process detector.  Returns decoder state
		result = m_pBpskdet->ProcBPSKDet(m_pBuffer,m_UseAFC);
		StatusChange(result);
// Process rig freq control if on
		if( m_UseAFC )
		{
			INT freqtmp;
			double doptmp;
			m_pBpskdet->GetRXFrequency(&freqtmp, &doptmp );
			m_RigCntrl.CalculateFrequencyControl(freqtmp, doptmp, result );
		}
		else
		{
			m_RigCntrl.ResetCtrl(m_pBpskdet);
		}
	}
	if(!m_UseInWave)
	{
		if(	m_pSoundDev )
		{
			m_pSoundDev->InClose();
		}
	}
	m_RigCntrl.Stop();
	m_ThreadActive = FALSE;
	StatusChange(DLLSTAT_OFF);
	DeleteResources();		//kill all resources
}


///////////////////////////////////////////////////////////////////////////
//  Free resources and exit process thread
//   Called by process thread itself to shut itself off if not signaled by
//  main program thread.
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::StopThread()
{
//  if any resources are open, kill em.
	m_pProcThread->m_bAutoDelete = TRUE;	//make thread clean up after
	m_pProcThread = NULL;  				// itself before it dies
	DeleteResources();
	m_DLLStatus = DLLSTAT_OFF;
	AfxEndThread(2); 
}

///////////////////////////////////////////////////////////////////////////
//   Called to delete any allocated resources in this class
///////////////////////////////////////////////////////////////////////////
void CIOCntrl::DeleteResources()
{
	m_ThreadActive = FALSE;
	if(	m_pSoundDev)
	{
		m_pSoundDev->InClose();
		delete m_pSoundDev;
		m_pSoundDev = NULL;
	}
	WaveDev.InClose();
	WaveDev.OutClose();
	if(m_pBuffer)
	{
		delete m_pBuffer;
		m_pBuffer = NULL;
	}
}

//////////////////////////////////////////////////////////////////////////
// Simple minded error message display for various soundcard errors
// Should be put in string table.
//////////////////////////////////////////////////////////////////////////
void CIOCntrl::ProcessError( INT ErrorCode )
{
	m_DLLErrStatus = ErrorCode;
	switch( ErrorCode )
	{
		case NO_ERRORS:
			m_ErrorMessage = _T("");
			break;
		case MEMORY_ERROR:
		case MMSYSERR_NOMEM:
			m_ErrorMessage = _T("Memory Allocation error");
			ErrorCode = DLLERR_MEMORY;
			break;
		case SOUNDIN_ERR_NOTOPEN:
			m_ErrorMessage = _T("Tried to read and soundcard is not open");
			ErrorCode = DLLERR_IN_NOTOPEN;
			break;
		case SOUNDIN_ERR_OVERFLOW:
			m_ErrorMessage = _T("Input buffers overflowed");
			ErrorCode = DLLERR_IN_OVERFLOW;
			break;
		case SOUNDIN_ERR_TIMEOUT:
			m_ErrorMessage = _T("Timed out waiting for input buffers");
			ErrorCode = DLLERR_IN_TIMEOUT;
			break;
		case SOUNDOUT_ERR_NOTOPEN:
			m_ErrorMessage = _T("Tried to write and soundcard is not open");
			ErrorCode = DLLERR_OUT_NOTOPEN;
			break;
		case SOUNDOUT_ERR_UNDERFLOW:
			m_ErrorMessage = _T("Output buffers Underflowed");
			ErrorCode = DLLERR_OUT_UNDERFLOW;
			break;
		case SOUNDOUT_ERR_TIMEOUT:
			m_ErrorMessage = _T("Timed out waiting for output buffers");
			ErrorCode = DLLERR_OUT_TIMEOUT;
			break;

		case WAVERR_BADFORMAT:
			m_ErrorMessage = _T("Wave file is not 8000Hz, Mono format");
			ErrorCode = DLLERR_WAV_BADFORMAT;
			break;
		case WAVERR_STILLPLAYING:
			m_ErrorMessage = _T("still something playing");
			ErrorCode = DLLERR_WAV_STILLPLAYING;
			break;
		case WAVERR_UNPREPARED:
			m_ErrorMessage = _T("header not prepared ");
			ErrorCode = DLLERR_WAV_UNPREPARED;
			break;
		case WAVERR_SYNC:
			m_ErrorMessage = _T("device is synchronous");
			ErrorCode = DLLERR_WAV_SYNC;
			break;

		case MMSYSERR_BADDEVICEID:
			m_ErrorMessage = _T("Bad Device ID\nIs Soundcard Present?");
			ErrorCode = DLLERR_SYS_BADDEVICEID;
			break;
		case MMSYSERR_NOTENABLED:
			m_ErrorMessage = _T("Driver failed enable");
			ErrorCode = DLLERR_SYS_NOTENABLED;
			break;
		case MMSYSERR_ALLOCATED:
			m_ErrorMessage = _T("Device already allocated");
			ErrorCode = DLLERR_SYS_ALLOCATED;
			break;
		case MMSYSERR_INVALHANDLE:
			m_ErrorMessage = _T("Device handle is invalid");
			ErrorCode = DLLERR_SYS_INVALHANDLE;
			break;
		case MMSYSERR_NODRIVER:
			m_ErrorMessage = _T("No device driver present");
			ErrorCode = DLLERR_SYS_NODRIVER;
			break;
		case MMSYSERR_NOTSUPPORTED:
			m_ErrorMessage = _T("Function isn't supported");
			ErrorCode = DLLERR_SYS_NOTSUPPORTED;
			break;
		case MMSYSERR_BADERRNUM:
			m_ErrorMessage = _T("Error value out of range");
			ErrorCode = DLLERR_SYS_BADERRNUM;
			break;
		case MMSYSERR_INVALFLAG: 
			m_ErrorMessage = _T("Invalid flag passed");
			ErrorCode = DLLERR_SYS_INVALFLAG;
			break;
		case MMSYSERR_INVALPARAM: 
			m_ErrorMessage = _T("Invalid parameter passed");
			ErrorCode = DLLERR_SYS_INVALPARAM;
			break;
		case MMSYSERR_HANDLEBUSY:
			m_ErrorMessage = _T("Handle being used");
			ErrorCode = DLLERR_SYS_HANDLEBUSY;
			break;
		case MMSYSERR_NODRIVERCB:
			m_ErrorMessage = _T("driver does not call DriverCallback");
			ErrorCode = DLLERR_SYS_NODRIVERCB;
			break;
		case MMSYSERR_INVALIDALIAS:
		case MMSYSERR_BADDB:
		case MMSYSERR_KEYNOTFOUND:
		case MMSYSERR_READERROR:
		case MMSYSERR_WRITEERROR:
		case MMSYSERR_DELETEERROR:
		case MMSYSERR_VALNOTFOUND:
			m_ErrorMessage = _T("Registry error");
			ErrorCode = DLLERR_SYS_REGISTRY;
			break;
		default:
			m_ErrorMessage = _T("Unknown Error");
			ErrorCode = DLLERR_UNKNOWN;
			break;
	}
	if(	m_ThreadActive )
		StatusChange(ErrorCode);
}

void CIOCntrl::StatusChange(INT code)
{
	if( code != m_DLLStatus )
	{
		m_DLLStatus = code;
		::PostMessage( m_hWnd, MSG_STATUSCHANGE, code, m_Progress);
	}
}








