// SpinButton.cpp : implementation file
//

#include "stdafx.h"
#include "SpinButton.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSpinButton

CSpinButton::CSpinButton()
{
	m_bDrawLockout = FALSE;
	m_bStayDown = FALSE;
}

CSpinButton::~CSpinButton()
{
}


BEGIN_MESSAGE_MAP(CSpinButton, CButton)
	//{{AFX_MSG_MAP(CSpinButton)
	ON_CONTROL_REFLECT(BN_CLICKED, OnClicked)
	ON_WM_LBUTTONDBLCLK()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSpinButton message handlers

void CSpinButton::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct) 
{
CPen* pOldPen;
	CRect rect (&(lpDrawItemStruct->rcItem));
	CDC *pDC = CDC::FromHandle(lpDrawItemStruct->hDC);
	UINT state = lpDrawItemStruct->itemState;
	UINT action = lpDrawItemStruct->itemAction;

	// Fill the entire button with the button brush 
	HBRUSH hbrText = GetSysColorBrush(COLOR_BTNTEXT);
	HBRUSH hbrFace = GetSysColorBrush(COLOR_3DFACE);
	HBRUSH hOldBrush = (HBRUSH) SelectObject(pDC->m_hDC, hbrFace);
 	FillRect(pDC->m_hDC, rect, hbrFace);
	CRect edgeRect(rect);
	if (state & ODS_FOCUS)
	{
		pDC->Rectangle(rect);
		edgeRect.InflateRect(-1, -1);
	}		

	if (state & ODS_SELECTED)
	{
		CPen pen(PS_SOLID, 2, GetSysColor(COLOR_3DSHADOW));
		pOldPen = pDC->SelectObject(&pen);
		pDC->Rectangle(edgeRect);
		pDC->SelectObject(pOldPen);
	}
	else
	{
		pDC->DrawEdge(edgeRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	}

	// Calculate the width and height of the triangle
	// based on the M character
	CSize szTri = pDC->GetOutputTextExtent("M", 1);

	// Calculate the width and height of the text
	CString text;
	GetWindowText(text);
	CSize szText = pDC->GetOutputTextExtent(text, text.GetLength());

	// Calculate the total size of the text + the triangle
	CSize szTotal(szText);
	szTotal.cx += szTri.cx;

	// change the width/height of the button rect
	CRect textRect = rect;
	textRect.InflateRect(-((rect.Width() - szTotal.cx)/2), 
		-((rect.Height() - szTotal.cy)/2));
	textRect.top--;

	CRect triRect = textRect;
	triRect.left = textRect.right - szTri.cx;
	triRect.InflateRect(0, -((textRect.Height() - szTri.cy)/2));

	// Draw the text
	pDC->SetBkMode(TRANSPARENT);
	if (state & ODS_SELECTED)
		textRect.OffsetRect(1, 1);
	pDC->DrawText(text, -1, textRect, DT_SINGLELINE | DT_CENTER);

	// Draw the triangle
	SelectObject(pDC->m_hDC, hbrText);
	CPen pen(PS_SOLID, 1, RGB(255,0,0) );
	pOldPen = pDC->SelectObject(&pen);
	CPoint ptTriL(rect.left + szTri.cx, textRect.top + 5);
	DrawTriangle(pDC, ptTriL, 13);
	CPoint ptTriR(rect.right - 2*szTri.cx, textRect.top + 5);
	DrawTriangle(pDC, ptTriR, 13);
	pDC->SelectObject(pOldPen);
	
	// Draw the focus rectangle if necessary
	if (state & ODS_FOCUS)
	{
		CRect focusRect(rect);
//		focusRect.InflateRect(-4, -4);
//		pDC->DrawFocusRect(focusRect);		
	}

	SelectObject(pDC->m_hDC, hOldBrush);
}


void CSpinButton::DrawTriangle(CDC *pDC, CPoint ptTopLeft, int nWidth)
{
	// Use odd widths for best results
	POINT ptArray[3];
	ptArray[0] = ptTopLeft;
	ptArray[1].x = ptTopLeft.x + nWidth - 1;
	ptArray[1].y = ptTopLeft.y;
	ptArray[2].x = ptTopLeft.x + nWidth / 2;
	ptArray[2].y = ptTopLeft.y + nWidth / 2;
	pDC->Polygon(ptArray, 3);
}


void CSpinButton::OnClicked() 
{
}


void CSpinButton::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	PostMessage(WM_LBUTTONDOWN, 
               MAKEWPARAM((WORD)point.x, (WORD)point.y),
               (DWORD)nFlags);

	CButton::OnLButtonDblClk(nFlags, point);
}
